#include "Transform.h"

/*!
    construct a unit transform
*/
Transform::Transform() 
{
    //for(int i=0;i<4;i++){
    //    m_mat(i,i)=1.0;
    //    m_inv_mat(i,i)=1.0;
    //}
    m_mat=Matrix4::unit();
    m_inv_mat=Matrix4::unit();
	m_scale=1.0;
}


Transform::Transform(const Transform& T) 
{
    m_mat=T.m_mat;
    m_inv_mat=T.m_inv_mat;
	m_scale=T.m_scale;
}
    
Transform::~Transform()
{}
  
// apply to 3-vector
Vector3 Transform::transform(const Vector3& v) const
{
    Vector3 res;
    std::array<double,4> tmp_v;
    std::array<double,4> tmp_res;
    

    // cartesian -> homogeneous coordinates
    for(int i=0;i<3;i++){
        tmp_v[i]=v[i];
    }
    tmp_v[3]=1.0;

    // matrix-vector multiply
    tmp_res.fill(0.0);
    for(int i=0;i<4;i++){
        for(int j=0;j<4;j++){
            tmp_res[i]+=m_mat(i,j)*tmp_v[j];
        }
    }
    
    // back to cartesian
    for(int i=0;i<3;i++){
        res[i]=tmp_res[i]/tmp_res[3];
    }
    return res;
}
    
Vector3 Transform::invtransform(const Vector3& v) const
{
    Vector3 res;
    std::array<double,4> tmp_v;
    std::array<double,4> tmp_res;
    

    // cartesian -> homogeneous coordinates
    for(int i=0;i<3;i++){
        tmp_v[i]=v[i];
    }
    tmp_v[3]=1.0;

    // matrix-vector multiply
    tmp_res.fill(0.0);
    for(int i=0;i<4;i++){
        for(int j=0;j<4;j++){
            tmp_res[i]+=m_inv_mat(i,j)*tmp_v[j];
        }
    }
    
    // back to cartesian
    for(int i=0;i<3;i++){
        res[i]=tmp_res[i]/tmp_res[3];
    }
    return res;
}
    
/*!
	Apply transformation to the normal of an object (plane,...)
	by multiplying with the transpose of the inverse transform
	_without_ the translation component. 

	\param n the normal vector
*/
Vector3 Transform::transform_n(const Vector3& n) const
{
	Vector3 res;
  

    // matrix-vector multiply
    for(int i=0;i<3;i++){
        for(int j=0;j<3;j++){
            res[i]+=m_inv_mat(j,i)*n[j];
        }
    }

    return res.unit(); // normals should stay unit length
}

/*!
	Apply the inverse transformation to the normal of an object (plane,...)
	by multiplying with the transpose of the _forward transform 
	without the translation component.
	
	\param n the normal vector
*/
Vector3 Transform::invtransform_n(const Vector3& n) const
{
	Vector3 res;
  
    // matrix-vector multiply
    for(int i=0;i<3;i++){
        for(int j=0;j<3;j++){
            res[i]+=m_mat(j,i)*n[j];
        }
    }
    
    return res;
}

// add in-place transform
void Transform::addTranslation(double x, double y, double z)
{
	// build translation matrix
 	Matrix4 mat=Matrix4::unit();
    mat(0,3)=x;
    mat(1,3)=y;
	mat(2,3)=z;
   
    // inverse translation matrix
	Matrix4 inv_mat=Matrix4::unit();
    inv_mat(0,3)=-1.0*x;
    inv_mat(1,3)=-1.0*y;
    inv_mat(2,3)=-1.0*z;
	// apply 
    m_mat=mat*m_mat;
    m_inv_mat=inv_mat*m_inv_mat;
    
}


void Transform::addRotationX(double phi)
{
	Matrix4 mat=Matrix4::unit();
    Matrix4 inv_mat=Matrix4::unit();
    // build rotation matrix
    mat(1,1)=cos(phi);
    mat(1,2)=-1.0*sin(phi);
    mat(2,1)=sin(phi);
    mat(2,2)=cos(phi);
   
	// build inverse rotation matrix
    inv_mat(1,1)=cos(phi);
    inv_mat(1,2)=sin(phi);
    inv_mat(2,1)=-1.0*sin(phi);
    inv_mat(2,2)=cos(phi);
       
    // apply 
    m_mat=mat*m_mat;
    m_inv_mat=m_inv_mat*inv_mat;
}

void Transform::addRotationY(double phi)
{
	Matrix4 mat=Matrix4::unit();
    Matrix4 inv_mat=Matrix4::unit();
    // build rotation matrix
    mat(0,0)=cos(phi);
    mat(0,2)=sin(phi);
    mat(2,0)=-1.0*sin(phi);
    mat(2,2)=cos(phi);
   
	// build inverse rotation matrix
    inv_mat(0,0)=cos(phi);
    inv_mat(0,2)=-1.0*sin(phi);
    inv_mat(2,0)=sin(phi);
    inv_mat(2,2)=cos(phi);
       
    // apply 
    m_mat=mat*m_mat;
    m_inv_mat=m_inv_mat*inv_mat;
}


void Transform::addRotationZ(double phi)
{
	Matrix4 mat=Matrix4::unit();
    Matrix4 inv_mat=Matrix4::unit();
    // build rotation matrix
    mat(0,0)=cos(phi);
    mat(0,1)=-1.0*sin(phi);
    mat(1,0)=sin(phi);
    mat(1,1)=cos(phi);
   
	// build inverse rotation matrix
    inv_mat(0,0)=cos(phi);
    inv_mat(0,1)=sin(phi);
    inv_mat(1,0)=-1.0*sin(phi);
    inv_mat(1,1)=cos(phi);
       
    // apply 
    m_mat=mat*m_mat;
    m_inv_mat=m_inv_mat*inv_mat;
}

/*!
	Add isotropic scaling to the translation
	
	\param s the scale factor
*/
void Transform::addScale(double s)
{
	Matrix4 mat=Matrix4::unit();
    Matrix4 inv_mat=Matrix4::unit();
	
	for(int i=0;i<3;i++){
		mat(i,i)=s;
		inv_mat(i,i)=1.0/s;
	}
	
	// apply 
    m_mat=mat*m_mat;
    m_inv_mat=m_inv_mat*inv_mat;
	m_scale*=s;
}

Transform Transform::translate(double x, double y, double z) const
{
    Transform res;
    Matrix4 mat=Matrix4::unit();
    Matrix4 inv_mat=Matrix4::unit();
    // build translation matrix
    mat(0,3)=x;
    mat(1,3)=y;
    mat(2,3)=z;
   
    // inverse translation matrix
    inv_mat(0,3)=-1.0*x;
    inv_mat(1,3)=-1.0*y;
    inv_mat(2,3)=-1.0*z;
   
    // apply 
    res.m_mat=mat*m_mat;
    res.m_inv_mat=m_inv_mat*inv_mat;
    
    return res;
}

Transform Transform::rotateX(double phi) const
{
    Transform res;
    
    Matrix4 mat=Matrix4::unit();
    Matrix4 inv_mat=Matrix4::unit();
    // build rotation matrix
    mat(1,1)=cos(phi);
    mat(1,2)=-1.0*sin(phi);
    mat(2,1)=sin(phi);
    mat(2,2)=cos(phi);
   
	// build inverse rotation matrix
    inv_mat(1,1)=cos(phi);
    inv_mat(1,2)=sin(phi);
    inv_mat(2,1)=-1.0*sin(phi);
    inv_mat(2,2)=cos(phi);
	
    // apply 
    res.m_mat=mat*m_mat;
    res.m_inv_mat=m_inv_mat*inv_mat;
    
    return res;
}
 
Transform Transform::rotateY(double phi) const
{
    Transform res;
 
	Matrix4 mat=Matrix4::unit();
    Matrix4 inv_mat=Matrix4::unit();
    // build rotation matrix
    mat(0,0)=cos(phi);
    mat(0,2)=sin(phi);
    mat(2,0)=-1.0*sin(phi);
    mat(2,2)=cos(phi);
   
	// build inverse rotation matrix
    inv_mat(0,0)=cos(phi);
    inv_mat(0,2)=-1.0*sin(phi);
    inv_mat(2,0)=sin(phi);
    inv_mat(2,2)=cos(phi);
       
    // apply 
    res.m_mat=mat*m_mat;
    res.m_inv_mat=m_inv_mat*inv_mat;
	
	return res;
}
	
Transform Transform::rotateZ(double phi) const
{
    Transform res;
    
    Matrix4 mat=Matrix4::unit();
    Matrix4 inv_mat=Matrix4::unit();
    // build rotation matrix
    mat(0,0)=cos(phi);
    mat(0,1)=-1.0*sin(phi);
    mat(1,0)=sin(phi);
    mat(1,1)=cos(phi);
   
	// build inverse rotation matrix
    inv_mat(0,0)=cos(phi);
    inv_mat(0,1)=sin(phi);
    inv_mat(1,0)=-1.0*sin(phi);
    inv_mat(1,1)=cos(phi);
       
    // apply 
    res.m_mat=mat*m_mat;
    res.m_inv_mat=m_inv_mat*inv_mat;
    
    return res;
}

/*!
	generate scaled transform
	
	\param s scale factor
*/
Transform Transform::scale(double s) const
{	
	Transform res;
    
    Matrix4 mat=Matrix4::unit();
    Matrix4 inv_mat=Matrix4::unit();
	
	for(int i=0;i<3;i++){
		mat(i,i)=s;
		inv_mat(i,i)=1.0/s;
	}
	
	// apply 
    res.m_mat=mat*m_mat;
    res.m_inv_mat=m_inv_mat*inv_mat;
	res.m_scale=s;
	
	return res;
}

/*
	multiply the transform with another one
*/
void Transform::addTransform(const Transform& T)
{
	// apply 
	m_scale=m_scale*T.m_scale;
    m_mat=T.m_mat*m_mat;
    m_inv_mat=m_inv_mat*T.m_inv_mat;
}

/*!
	Extract the scale factor contained in the transformation.
	Spits an error message if the scaling in non-isotropic by 
	a factor of >1e-6
*/
double Transform::getScale() const
{
	return m_scale;
}



std::ostream& operator << (std::ostream& ost,const Transform& T)
{
    ost << T.m_mat;
        
    return ost;
}
