#include "Sphere3DFitterNR.h"

// --- project includes ---
#include "sphere_fitting/utils/mat4.h"

/*! 
	contructor

	\param GO1 1st geometric object to which the sphere is fitted
	\param GO2 
	\param GO3 
	\param GO4
*/
Sphere3DFitterNR::Sphere3DFitterNR(const AGeometricObject* GO1,
								   const AGeometricObject* GO2,
								   const AGeometricObject* GO3,
								   const AGeometricObject* GO4) : m_x(0.0,4), m_dist(0.0,4)
{
	m_GO1=GO1;
	m_GO2=GO2;
	m_GO3=GO3;
	m_GO4=GO4;
}

/*!
	in-place update of current distance vector based on new position & radius
*/
void Sphere3DFitterNR::update_distances()
{
	Vector3 p=Vector3(m_x[0],m_x[1],m_x[2]);
	m_dist[0]=m_GO1->getDist(p)-m_x[3];
	m_dist[1]=m_GO2->getDist(p)-m_x[3];
	m_dist[2]=m_GO3->getDist(p)-m_x[3];
	m_dist[3]=m_GO4->getDist(p)-m_x[3];
}

/*!
	in-place update of current jacobian (distance derivatives) based on new position & radius
*/
void Sphere3DFitterNR::update_jacobian()
{
	Vector3 p=Vector3(m_x[0],m_x[1],m_x[2]);
	std::array<double,3> dd;
	
	dd=m_GO1->getDistDeriv(p);
	for(unsigned int i=0;i<3;i++){m_jacobian(0,i)=dd[i];}
	dd=m_GO2->getDistDeriv(p);
	for(unsigned int i=0;i<3;i++){m_jacobian(1,i)=dd[i];}
	dd=m_GO3->getDistDeriv(p);
	for(unsigned int i=0;i<3;i++){m_jacobian(2,i)=dd[i];}
	dd=m_GO4->getDistDeriv(p);
	for(unsigned int i=0;i<3;i++){m_jacobian(3,i)=dd[i];}
	for(unsigned int j=0;j<4;j++){m_jacobian(j,3)=-1.0;}
	
}

/*!
	return the 2-norm of the current distance (misfit) vector
*/
double Sphere3DFitterNR::dist_norm() const
{
	return sqrt((m_dist[0]*m_dist[0])+(m_dist[1]*m_dist[1])+(m_dist[2]*m_dist[2])+(m_dist[3]*m_dist[3]));
}

/*!
	Fit sphere to 4 geometric objects using a Newton-Raphson algorithm
	
	\param spos starting value for the N-R optimizer
	\param max_iter maximuma number of iterations to catch if the optimizer is "stuck", for example in a secondary minimum
	\param prec required precision for the convergence of the optimizer  
*/
Sphere Sphere3DFitterNR::calc_fit(const Vector3& spos,int max_iter,double prec)
{
	Sphere res;
	bool solution_valid=false;
	
	m_x[0]=spos.x();
	m_x[1]=spos.y();
	m_x[2]=spos.z();
	
	update_distances();
	unsigned int i=0;
	while ((i< max_iter) && (dist_norm()>prec)){
		update_jacobian();
		auto dx=m_jacobian.solve_destructive(m_dist,prec);
		if (dx.valid){
			m_x=m_x-dx.data;
			update_distances();
			solution_valid=true;
			i++;
		} else {
			i=max_iter+1;
			solution_valid=false;
		}
	}	
	if (solution_valid){
		Vector3 c=Vector3(m_x[0],m_x[1],m_x[2]);
		double rad=m_x[3]+m_dist.min();
		res=Sphere(c,rad-prec/2.0);
	}
	return res;
}

std::ostream& operator << (std::ostream& ost,const Sphere3DFitterNR& S)
{
	ost << "Sphere3DFitterNR" << std::endl;
	ost << "x: " << S.m_x[0] << " , " << S.m_x[1] << " , " << S.m_x[2] << " r: " << S.m_x[3] << "\n" ;
	ost << "dist: " << S.m_dist[0] << " , " << S.m_dist[1] << " , " << S.m_dist[2] << " , " << S.m_dist[3] << "\n" ;
	ost << "jacobian: \n" << S.m_jacobian << std::endl;
	return ost;
}


					 
