/////////////////////////////////////////////////////////////
//                                                         //
// Copyright (c) 2007-2017 by The University of Queensland //
// Centre for Geoscience Computing                         //
// http://earth.uq.edu.au/centre-geoscience-computing      //
//                                                         //
// Primary Business: Brisbane, Queensland, Australia       //
// Licensed under the Open Software License version 3.0    //
// http://www.apache.org/licenses/LICENSE-2.0              //
//                                                         //
/////////////////////////////////////////////////////////////

#include "MeshVolume.h"

// --- Project includes ---
#include "geometry/Triangle3D.h"

//-- system includes --
#include <cstdlib>

using std::rand;

// --- STL includes ---
#include <utility>
#include <set>
#include <limits>
#include <algorithm> // for std::max

// --- I/O includes
#include <iostream>

using std::make_pair;

/*!
  helper function, return random value between min & max
*/
double MeshVolume::m_random(double imin,double imax) const
{
  return imin+((imax-imin)*((double)(rand())/(double)(RAND_MAX)));
}

/*!
  Constructor
*/
MeshVolume::MeshVolume() : m_has_grid(false)
{}

/*!
  Constructor

  \param tp the triangle set
*/
MeshVolume::MeshVolume(const TriPatchSet& tp) 
{
  m_mesh=tp;
  m_MinPoint=m_mesh.getBoundingBox().first;
  m_MaxPoint=m_mesh.getBoundingBox().second;
  m_DistPoint=m_MinPoint-Vector3(1.0,1.0,1.0);
  if(m_mesh.size() > 20) {
	setup_accel_grid(); 
  } else {
	m_has_grid=false;
  }
}

/*!
	Copy constructor
*/
MeshVolume::MeshVolume(const MeshVolume& MV): m_mesh(MV.m_mesh), m_MinPoint(MV.m_MinPoint), m_MaxPoint(MV.m_MaxPoint), m_DistPoint(MV.m_DistPoint)
{
	m_has_grid=MV.m_has_grid;
	if(m_has_grid){
		setup_accel_grid();
	}
}

/*!
	Copy contructor with additional flag if acceleration grid is copied or not. 
*/
MeshVolume::MeshVolume(const MeshVolume& MV, bool copy_grid): m_mesh(MV.m_mesh), m_MinPoint(MV.m_MinPoint), m_MaxPoint(MV.m_MaxPoint), m_DistPoint(MV.m_DistPoint)
{
	if(copy_grid){
		m_has_grid=MV.m_has_grid;
	} else {
		m_has_grid=false;
	}
	if(m_has_grid){
		setup_accel_grid();
	}
}

/*!
  get bounding box
*/
pair<Vector3,Vector3> MeshVolume::getBoundingBox() 
{
  return m_mesh.getBoundingBox();
}

/*! 
  Get a point inside the volume. The argument is ignored.

  \warning if the mesh is incorrect the function might loop forever
*/
Vector3 MeshVolume::getAPoint(int) const
{
  Vector3 res;
  double px,py,pz;
 
  do{
    px=m_random(m_MinPoint.x(),m_MaxPoint.x());
    py=m_random(m_MinPoint.y(),m_MaxPoint.y());
    pz=m_random(m_MinPoint.z(),m_MaxPoint.z());
    res=Vector3(px,py,pz);
  } while (!isIn(res));

  return res;
}

/*!
  returns a map of the closest mesh facets and their distances to a given point

  \param P the point
  \param nmax the maximum number of objects to be returned
*/ 
const map<double,const AGeometricObject*> MeshVolume::getClosestObjects(const Vector3& P,int nmax) const
{
	map<double,const AGeometricObject*> res;

	// std::cout << "MeshVolume::getClosestObjects( " << P << " , " << nmax << " )" << std::endl;

	if(m_has_grid){ // check grid 
		// unsigned int n_tris_checked=0; // debug counter
		int srange=0;
		double dist4=std::numeric_limits<double>::max(); // set to max. value
		int i0,j0,k0;
		
		std::tie(i0,j0,k0)=get_grid_index(P); // axis indices of the starting cell
		// std::cout << "starting cell: " << i0 << " , " << j0 << " , " << k0 << std::endl;
		// "inside-out" search until either:
		// - 4 objects found and no further object with smaller distance can be in current search range
		// - search range exceeds grid size
		while(((res.size()<4) || (dist4>double(srange-1)*m_grid_spacing)) && (srange<m_nmax)){ 
			std::set<int> t_idx_set=std::set<int>();
			for(int i=-1*srange;i<srange+1;i++){
				for(int j=-1*srange;j<srange+1;j++){
					for(int k=-1*srange;k<srange+1;k++){
						if(std::max({abs(i),abs(j),abs(k)}) >= srange){ 
							int i1=i0+i;
							int j1=j0+j;
							int k1=k0+k;
							// std::cout << "offset : " << i << " , " << j << " , " << k << " cell : " << i1 << " , " << j1 << " , " << k1 << std::endl;
							if ((i1>=0) && (j1>=0) && (k1>=0) && (i1<m_nx) && (j1<m_ny) && (k1<m_nz)){ // check that we are actually inside the grid
								int idx=make_index(i1,j1,k1);
								// std::cout << "triangles in cell: " << m_grid[idx].size() << std::endl;
								for(auto t: m_grid[idx]){ // get all triangle indices in the voxel
									t_idx_set.insert(t); // insert into set -> ensures uniqeness
									//std::cout << " adding triangle: " << t << std::endl;
								}
							}
						}
					}
				}
			}
			// process the triangles found
			// std::cout << "checking " << t_idx_set.size() << " triangles at srange = " << srange << std::endl;
			for(auto t:t_idx_set){ 
				double ndist=m_mesh[t].getDist(P);
				res[ndist]=&(m_mesh[t]);
				// std::cout << "dist to triangle " << t << " = " << ndist << " objects in res: " <<  res.size() << std::endl;
				if(res.size()>4){
					auto last=res.rbegin();
					res.erase(last->first); // remove last element
					dist4=res.rbegin()->first;
					// std::cout << "dist4 = " << dist4 << std::endl;
				}
				// n_tris_checked++;
			} 
			srange++;
        }
		// std::cout << "total triangles checked: " << n_tris_checked << std::endl; 	
	} else { // check against all triangles
		for(vector<Triangle3D>::const_iterator iter=m_mesh.triangles_begin();
			iter!=m_mesh.triangles_end();
			iter++){
			double ndist=iter->getDist(P);
			res[ndist]=&(*iter);
		}
	}
	return res;  
}

/*!
  Check if a point is inside the volume.  Works by checking if a 
  line between the point and a point outside crosses
  an odd number of mesh facets.

  \param P the point
*/
bool MeshVolume::isIn(const Vector3& P) const
{
	int cross_count=0;
	if(m_has_grid){ // use grid accelerated search
		int i0,j0,k0;
		std::set<int> t_idx_set;
		
		Vector3 dpoint(m_MaxPoint.x()+1.0,P.y(),P.z()); // define direction of ray to be parallel to +x to simplify voxel walking
		std::tie(i0,j0,k0)=get_grid_index(P); // axis indices of the starting cell
		if((i0>=0) && (j0>=0) && (k0>=0) && (i0<m_nx) && (j0<m_ny) && (k0<m_nz)) {
			int idx0=make_index(i0,j0,k0);
			int idx1=idx0+(m_nx-i0);
			// std::cout << "P: " << P << "  i,j,k: " << i0 << " , " << j0 << " , " << k0 << std::endl;
			// std::cout << "idx0,idx1: " << idx0 << " , " << idx1 << std::endl;
			for(int i=idx0;i<idx1;i++){ // walk voxels towards +x 
				for(auto t: m_grid[i]){ // get all triangle indices in the voxel
					t_idx_set.insert(t); // insert into set -> ensures uniqeness
				}
			} 
			// std::cout << "nt: " << t_idx_set.size() << std::endl;
			// check crossing for collected triangles
			for(auto t:t_idx_set){
				if(m_mesh[t].crosses(P,dpoint)) cross_count++;
			} 
		}
	} else { // check against all triangles
		for(vector<Triangle3D>::const_iterator iter=m_mesh.triangles_begin();
			iter!=m_mesh.triangles_end();
			iter++){
			if(iter->crosses(P,m_DistPoint)) cross_count++;
		}
	}
    return (cross_count%2)==1;
}

/*!
  Check if a sphere is fully inside the volume. Works by testing if center of the 
  sphere is inside and sphere doesn't intersect a mesh facet.

  \param S the sphere
*/
bool MeshVolume::isIn(const Sphere& S)
{
	bool res=isIn(S.Center());	
	if(m_has_grid){
		int i0,j0,k0,i1,j1,k1;
		std::set<int> t_idx_set;
		
		std::tie(i0,j0,k0)=get_grid_index(S.Center()-Vector3(S.Radius())); // axis indices of the min. cell intersecting sphere
		std::tie(i1,j1,k1)=get_grid_index(S.Center()+Vector3(S.Radius())); // axis indices of the min. cell intersecting sphere
		
		for(int i=std::max(i0,0);i<=std::min(i1,m_nx-1);i++) {
			for(int j=std::max(j0,0);j<=std::min(j1,m_ny-1);j++) {
				for(int k=std::max(k0,0);k<=std::min(k1,m_nz-1);k++) {
					int idx=make_index(i,j,k);
					for(auto t: m_grid[idx]){ // get all triangle indices in the voxel
						t_idx_set.insert(t); // insert into set -> ensures uniqeness
					}
				}
			}
		}
		// std::cout << "nt: " << t_idx_set.size() << std::endl;
		// check crossing for collected triangles
		for(auto t:t_idx_set){
			res=res && (m_mesh[t].getDist(S.Center())>S.Radius()) ;
		} 
  } else {
	  vector<Triangle3D>::const_iterator iter=m_mesh.triangles_begin();
	  while((iter!=m_mesh.triangles_end()) && res){
		res=(iter->getDist(S.Center())>S.Radius());
		iter++;
	  }
  }
  
  return res;
}

/*!
  Check if a sphere is fully outside the volume. Works by testing if center of the 
  sphere is not inside and sphere doesn't intersect a mesh facet.

  \param S the sphere
*/
bool MeshVolume::isFullyOutside(const Sphere& S)
{
  bool res=!isIn(S.Center());

  vector<Triangle3D>::const_iterator iter=m_mesh.triangles_begin();
  while((iter!=m_mesh.triangles_end()) && res){
    res=(iter->getDist(S.Center())>S.Radius());
    iter++;
  }

  return res;
}

/*!
	get grid index for a point
	
	\param p the point
	\warning the existence of a grid is not checked
*/
std::tuple<int,int,int> MeshVolume::get_grid_index(const Vector3& p) const
{
	Vector3 p1=p-m_MinPoint;

	int i=int(std::floor(p1.x()/m_grid_spacing));
	int j=int(std::floor(p1.y()/m_grid_spacing));
	int k=int(std::floor(p1.z()/m_grid_spacing));

	return std::make_tuple(i,j,k);
}

/*
	get linear index in m_grid from axis indices
	
	\param i x-axis index
	\param j y-axis index
	\param k z-axis index
*/
int MeshVolume::make_index(int i, int j, int k) const
{
	return m_nx*m_ny*k+m_nx*j+i;
}


/*!
	Setup the grid mapping of the triangles used accelerate
	the ray-triangle intersection in isIn() and the search
	for the closest objects.
*/
void MeshVolume::setup_accel_grid()
{
	// calculate the number of grid cells - should be roughly the same as number of triangles
	double vol=(m_MaxPoint.x()-m_MinPoint.x())*(m_MaxPoint.y()-m_MinPoint.y())*(m_MaxPoint.z()-m_MinPoint.z());
	double cell_vol=vol/float(m_mesh.size());
	m_grid_spacing=std::pow(cell_vol,1.0/3.0);
	m_nx=int (std::ceil((m_MaxPoint.x()-m_MinPoint.x())/m_grid_spacing));
	m_ny=int (std::ceil((m_MaxPoint.y()-m_MinPoint.y())/m_grid_spacing));
	m_nz=int(std::ceil((m_MaxPoint.z()-m_MinPoint.z())/m_grid_spacing));
	m_nmax=std::max({m_nx,m_ny,m_nz});
	unsigned int ncells=m_nx*m_ny*m_nz;
		
	// debug output
	// std::cout << "volume: " << vol << " triangles: " << m_mesh.size() << " cell volume: " << cell_vol << std::endl; 
	// std::cout << "grid spacing: " << m_grid_spacing << std::endl;
	std::cout << "grid dimension: " << m_nx << " x " << m_ny << " x " << m_nz << " nmax " << m_nmax << " cells: " << ncells << std::endl;   
	
	// allocate grid
	m_grid.resize(ncells);
	
	// populate grid with triangles
	int t_idx=0;
	for(vector<Triangle3D>::const_iterator iter=m_mesh.triangles_begin();
		iter!=m_mesh.triangles_end();
		iter++){
		int i0,j0,k0,i1,j1,k1;  // min/max cell indices per axis
		
		std::tie(i0,j0,k0)=get_grid_index(iter->getMinPoint());
		std::tie(i1,j1,k1)=get_grid_index(iter->getMaxPoint());
		
		// debug output
		// std::cout << t_idx << " : " << *iter << std::endl;
		// std::cout << "i: " << i0 << " - " << i1 <<  " j :" << j0 << " - " << j1 <<  " k :" << k0 << " - " << k1 << std::endl; 
		
		for(int i=i0; i<=i1; i++){
			for(int j=j0; j<=j1; j++){
				for(int k=k0;k<=k1; k++){
					int idx=make_index(i,j,k);
					m_grid[idx].push_back(t_idx);
				}
			}
		}
		t_idx++;
	}
	m_has_grid=true;

	// print_grid();
}

/*!
	Write the acceleration grid to stdout, if it exists.
	Mainly for debugging purposes.
*/
void MeshVolume::print_grid()
{
	if (m_has_grid) {
		std::cout << "grid" << std::endl;
		std::cout << "grid dimension: " << m_nx << " x " << m_ny << " x " << m_nz << std::endl;   
		
		for(int i=0; i<m_nx; i++){
			for(int j=0; j<m_ny; j++){
				for(int k=0;k<m_nz; k++){
					int idx=make_index(i,j,k);
					std::cout << i << " " << j << " " << k << " | " << idx << std::endl;
					for(auto t: m_grid[idx]){
						std::cout << t << " ";
					}
					std::cout << std::endl;
				}
			}
		}				
	} else {
		std::cout << "no grid" << std::endl;
	}
}

/*!
	force the use of the grid acceleration on/off
	
	\param use_grid true: on, false: off
*/
void MeshVolume::setGridAccel(bool use_grid)
{
	if(use_grid && (!m_has_grid)){
		setup_accel_grid();
	}
	if (!use_grid) {
		m_has_grid=false;
	}
}

