#include "TransformableBox3D.h"

// --- STL includes ---
#include <utility>
#include <list>
#include <algorithm>

using std::make_pair;
using std::list;

//--- IO includes ---
#include <iostream>

TransformableBoxWithPlanes3D::TransformableBoxWithPlanes3D() : BoxWithPlanes3D(), m_transform()
{}

TransformableBoxWithPlanes3D::TransformableBoxWithPlanes3D(const Vector3& pmin, const Vector3& pmax): BoxWithPlanes3D(pmin,pmax), m_transform()
{}

/*!
	Copy contructor
	
	\param B the box to be copied
*/
TransformableBoxWithPlanes3D::TransformableBoxWithPlanes3D(const TransformableBoxWithPlanes3D& B): BoxWithPlanes3D(B), m_transform(B.m_transform)
{
}


/*!
  get point inside the box. The Argument is ignored
*/
Vector3 TransformableBoxWithPlanes3D::getAPoint(int) const
{
	Vector3 P=BoxWithPlanes3D::getAPoint(0); // get untransformed point from base class 
	Vector3 Pt=m_transform.transform(P); // forward transform the point 
	
	return Pt;
}

/*!
  check if point is inside

  \param p the point
*/
bool TransformableBoxWithPlanes3D::isIn(const Vector3& p) const
{
	Vector3 pt=m_transform.invtransform(p); // backward transform point
	
	return BoxWithPlanes3D::isIn(pt); // check against untransformed box
}

/*!
  check if a Sphere is inside and doesn't intersect any planes

  \param S the Sphere
*/
bool TransformableBoxWithPlanes3D::isIn(const Sphere& S)
{
	double r=S.Radius()/m_transform.getScale();
	Vector3 p=S.Center();
	Vector3 pt=m_transform.invtransform(p); // backward transform sphere center
	
	// check transformed sphere against untransformed box
	bool inside=(pt.X()>m_pmin.X()+r) && (pt.X()<m_pmax.X()-r) && 
		(pt.Y()>m_pmin.Y()+r) && (pt.Y()<m_pmax.Y()-r) &&
		(pt.Z()>m_pmin.Z()+r) && (pt.Z()<m_pmax.Z()-r);
	  
	// check untransformed Sphere against planes  
	vector<Plane>::iterator iter=m_planes.begin();	  
	double dist=2*r;
	while((iter!=m_planes.end()) && (dist>r)){
		dist=iter->getDist(p);
		iter++;
	}

	return inside && (dist>r);  
}

/*
  check if a Sphere is fully outside the box
 \param S the Sphere
*/
bool TransformableBoxWithPlanes3D::isFullyOutside(const Sphere& S)
{
	double r=S.Radius();
	Vector3 p=S.Center();

	Vector3 pt=m_transform.invtransform(p); // backward transform sphere center
	Sphere St(pt,r); // contruct transformed sphere
	
	return BoxWithPlanes3D::isFullyOutside(St); // check against untransformed box
}


/*!
	Add a transformation to the Box, i.e. the current
	transformation matrix is multiplied by the supplied transformation.
	
	\param T the new transformation
*/
void TransformableBoxWithPlanes3D::addTransformation(const Transform& T)
{
	m_transform.addTransform(T); // update volume transform
	// update transforms of planes
	for(auto &iter : m_planes){ // mutable -> force reference (auto &)
		iter.addTransformation(T);
	}
}

/*!
  get bounding box
*/
pair<Vector3,Vector3> TransformableBoxWithPlanes3D::getBoundingBox()
{
	// can't rely on pmin/pmax -> transform _all_ corners & get min/max per axis
	Vector3 dp=m_pmax-m_pmin;
	Vector3 dx(dp.x(),0.0,0.0);
	Vector3 dy(0.0,dp.y(),0.0);
	Vector3 dz(0.0,0.0,dp.z());
	
	
	// temp storage for corner coordinates
	list<double> xlist;
	list<double> ylist;
	list<double> zlist;
	
	// walk trough all corners
	for(int i=0;i<2;i++){
		for(int j=0;j<2;j++){
			for(int k=0;k<2;k++){
				Vector3 corner=m_pmin+double(i)*dx+double(j)*dy+double(k)*dz;
				//std::cout << "corner:" << corner << std::endl;
				Vector3 c_trans=m_transform.transform(corner);
				// std::cout << "c_trans:" << c_trans << std::endl;
				xlist.push_back(c_trans.x());
				ylist.push_back(c_trans.y());
				zlist.push_back(c_trans.z());
			}
		}
	}
	
	// get per-axis min/max
	auto xrange=std::minmax_element(xlist.begin(),xlist.end());
	auto yrange=std::minmax_element(ylist.begin(),ylist.end());
	auto zrange=std::minmax_element(zlist.begin(),zlist.end());
	
	// construct bbx corners
	Vector3 bbx_min(*(xrange.first),*(yrange.first),*(zrange.first));
	Vector3 bbx_max(*(xrange.second),*(yrange.second),*(zrange.second));
	
	return make_pair(bbx_min,bbx_max);
}

/*!
	Generate a new, transformed volume. The original volume remains unchanged.

	\param T the transform
*/
TransformableBoxWithPlanes3D TransformableBoxWithPlanes3D::transform(const Transform& T) const
{
	TransformableBoxWithPlanes3D res(*this);
	res.addTransformation(T);
	
	return res;
}


ostream& operator <<(ostream& ost, const TransformableBoxWithPlanes3D& T)
{
	ost << "TransformableBoxWithPlanes3D: " << std::endl;
	ost << "t: " << T.m_transform << std::endl;
	ost << "pmin/pmax (orig): " << T.m_pmin << " - " << T.m_pmax << std::endl; 
	ost << "pmin/pmax (transformed): " << T.m_transform.transform(T.m_pmin) << " - " << T.m_transform.transform(T.m_pmax) << std::endl; 
	ost << "Planes:" << std::endl;
	for (auto iter : T.m_planes){
		ost << iter << std::endl;
	}
	
	return ost;
}
